"""Manage FBC curation results."""
import json
import logging
from pathlib import Path
from typing import Dict, List

import numpy as np
import pandas as pd

from fbc_curation.constants import CuratorConstants


logger = logging.getLogger(__name__)


class CuratorResults:
    """Class for working with fbc curation results.

    CuratorResults can either be created by a Curator or read from file.
    """

    def __init__(
        self,
        metadata: dict,
        objective_id: str,
        objective: pd.DataFrame,
        fva: pd.DataFrame,
        gene_deletion: pd.DataFrame,
        reaction_deletion: pd.DataFrame,
        num_decimals: int = None,
    ):
        """Create instance."""
        self.metadata = metadata
        self.objective_id = objective_id
        self.objective = objective
        self.fva = fva
        self.gene_deletion = gene_deletion
        self.reaction_deletion = reaction_deletion

        if not num_decimals:
            num_decimals = CuratorConstants.NUM_DECIMALS
        self.num_decimals = num_decimals

        # FIXME: processing must be done on creating the files ?!
        # round and sort objective value
        for key in ["value"]:
            self.objective[key] = self.objective[key].apply(self._round)
        self.objective.sort_values(by=["objective"], inplace=True)

        # round and sort fva
        for key in ["flux", "minimum", "maximum"]:
            self.fva[key] = self.fva[key].apply(self._round)
        self.fva.sort_values(by=["reaction"], inplace=True)
        self.fva.index = range(len(self.fva))

        # round and sort gene_deletion
        for key in ["value"]:
            self.gene_deletion[key] = self.gene_deletion[key].apply(self._round)
        self.gene_deletion.sort_values(by=["gene"], inplace=True)
        self.gene_deletion.index = range(len(self.gene_deletion))

        # round and sort reaction deletion
        for key in ["value"]:
            self.reaction_deletion[key] = self.reaction_deletion[key].apply(self._round)
        self.reaction_deletion.sort_values(by=["reaction"], inplace=True)
        self.reaction_deletion.index = range(len(self.reaction_deletion))

        # validate
        self.validate()

    def _round(self, x):
        """Round the float and sets small values positive.

        Ensuring positivity removes -0.0, 0.0 changes to files.
        """
        if x == CuratorConstants.VALUE_INFEASIBLE:
            return x
        else:
            x = round(x, self.num_decimals)
            if abs(x) < 1e-10:
                x = abs(x)
            return x

    def write_results(self, path_out: Path):
        """Write results to path."""
        if not path_out.exists():
            logger.warning(f"Creating results path: {path_out}")
            path_out.mkdir(parents=True)

        # write metadata file
        with open(path_out / CuratorConstants.METADATA_FILENAME, "w") as f_json:
            json.dump(self.metadata, fp=f_json, indent=2)

        # write reference files
        for filename, df in dict(
            zip(
                [
                    CuratorConstants.OBJECTIVE_FILENAME,
                    CuratorConstants.FVA_FILENAME,
                    CuratorConstants.GENE_DELETION_FILENAME,
                    CuratorConstants.REACTION_DELETION_FILENAME,
                ],
                [self.objective, self.fva, self.gene_deletion, self.reaction_deletion],
            )
        ).items():
            print(f"-> {path_out / filename}")
            df.to_csv(path_out / filename, sep="\t", index=False)
            # df.to_json(path_out / filename, sep="\t", index=False)

    @classmethod
    def read_results(cls, path_in: Path):
        """Read fbc curation files from given directory."""
        path_metadata = path_in / CuratorConstants.METADATA_FILENAME
        path_objective = path_in / CuratorConstants.OBJECTIVE_FILENAME
        path_fva = path_in / CuratorConstants.FVA_FILENAME
        path_gene_deletion = path_in / CuratorConstants.GENE_DELETION_FILENAME
        path_reaction_deletion = path_in / CuratorConstants.REACTION_DELETION_FILENAME
        df_dict = dict()

        for k, path in enumerate(
            [path_objective, path_fva, path_gene_deletion, path_reaction_deletion]
        ):
            if not path_objective.exists():
                logger.error(f"Required file for fbc curation does not exist: '{path}'")
            else:
                df_dict[CuratorConstants.KEYS[k]] = pd.read_csv(path, sep="\t")

        with open(path_metadata, "r") as f_json:
            df_dict[CuratorConstants.METADATA_KEY] = json.load(fp=f_json)
        objective_id = df_dict["objective"].objective.values[0]

        return CuratorResults(objective_id=objective_id, **df_dict)

    @staticmethod
    def compare(results: Dict[str, "CuratorResults"]) -> bool:
        """Compare results against each other.

        Returns True of all results are identical.
        """
        curator_keys = list(results.keys())
        curator_results = list(results.values())
        keys = list(results.keys())
        num_res = len(curator_results)
        # only comparing comparison between two data frames

        print("=" * 40)
        print("Comparison of results")
        print("=" * 40)
        all_equal = True

        for key in CuratorConstants.KEYS:
            mat_equal = np.ndarray(shape=(num_res, num_res))
            for p, res1 in enumerate(curator_results):
                for q, res2 in enumerate(curator_results):
                    df1 = getattr(res1, key)
                    df2 = getattr(res2, key)
                    if df1.equals(df2):
                        equal = 1
                    else:
                        equal = 0
                    mat_equal[p, q] = equal

                    if equal == 0:
                        print(f"difference: '{curator_keys[p]}' vs '{curator_keys[q]}'")
                        CuratorResults.analyse_df_differende(df1, df2)

            df_equal = pd.DataFrame(
                mat_equal, columns=list(keys), index=list(keys), dtype=int
            )
            print(f"--- {key} ---")
            print(df_equal)
            all_equal = (
                all_equal and np.sum(np.sum(df_equal.values)) == num_res * num_res
            )

        print("=" * 40)
        print(f"Equal: {all_equal}")
        print("=" * 40)
        return all_equal

    @staticmethod
    def analyse_df_differende(df1: pd.DataFrame, df2: pd.DataFrame):
        """Analyse DataFrame difference."""
        df_diff = pd.concat([df1, df2]).drop_duplicates(keep=False)
        print(df_diff)

    def validate(self) -> bool:
        """Validate results."""
        valid_objective = self.validate_objective()
        valid_fva = self.validate_fva()
        valid_gene_deletion = self.validate_gene_deletion()
        valid_reaction_deletion = self.validate_reaction_deletion()
        return (
            valid_objective
            and valid_fva
            and valid_gene_deletion
            and valid_reaction_deletion
        )

    def validate_objective(self) -> bool:
        """Validate objective."""
        return CuratorResults._validate_df(
            self.objective,
            name=CuratorConstants.OBJECTIVE_KEY,
            fields=CuratorConstants.OBJECTIVE_FIELDS,
        )

    def validate_fva(self) -> bool:
        """Validate FVA."""
        return CuratorResults._validate_df(
            self.fva, name=CuratorConstants.FVA_KEY, fields=CuratorConstants.FVA_FIELDS
        )

    def validate_gene_deletion(self) -> bool:
        """Validate gene deletion."""
        return CuratorResults._validate_df(
            self.gene_deletion,
            name=CuratorConstants.GENE_DELETION_KEY,
            fields=CuratorConstants.GENE_DELETION_FIELDS,
        )

    def validate_reaction_deletion(self) -> bool:
        """Validate reaction deletion."""
        return CuratorResults._validate_df(
            self.reaction_deletion,
            name=CuratorConstants.REACTION_DELETION_KEY,
            fields=CuratorConstants.REACTION_DELETION_FIELDS,
        )

    @staticmethod
    def _validate_df(df: pd.DataFrame, name: str, fields: List[str]) -> bool:
        valid = True
        if not isinstance(df, pd.DataFrame):
            logger.error(f"'{name}': Must be 'pd.DataFrame', but type '{type(df)}'.")
            valid = False
        if df.empty:
            logger.error(f"'{name}': Can not be empty.")
            valid = False
        if len(df.columns) != len(fields):
            logger.error(
                f"'{name}': Incorrect number of columns: "
                f"'{len(df.columns)} != {len(fields)}'."
            )
            valid = False
        for field in fields:
            if field not in df.columns:
                logger.error(f"'{name}': Missing field '{field}'")
                valid = False
        for k, field in enumerate(fields):
            if not df.columns[k] == field:
                logger.error(
                    f"'{name}': Field at position '{k}' must be {field}', "
                    f"but is '{df.columns[k]}'."
                )
                valid = False

        for status_code in df.status.unique():
            if status_code not in CuratorConstants.STATUS_CODES:
                logger.error(f"'{name}': Incorrect status code: '{status_code}'.")

        if name == CuratorConstants.OBJECTIVE_KEY:
            obj_value = df["value"].values[0]
            if not obj_value > 0:
                logger.error(
                    f"'{name}': objective value must be > 0, but '{obj_value}'."
                )
                valid = False

        if valid:
            logger.info(f"'{name}': is VALID")
        else:
            logger.error(f"'{name}': is INVALID")

        return valid



"""Module creates FBC curation files."""
import logging
from pathlib import Path

from pyfiglet import Figlet

from fbc_curation import __citation__, __version__
from fbc_curation.curator import Curator, CuratorResults


logger = logging.getLogger(__name__)


def main():
    """Entry point which runs fbc_curation script.

    Example:
        python curation.py --model ../examples/models/e_coli_core.xml
          --path ../examples/results/e_coli_core
        fbc_curation --model examples/models/e_coli_core.xml
          --path examples/results/e_coli_core
    """

    import optparse
    import sys

    parser = optparse.OptionParser()
    parser.add_option(
        "-m",
        "--model",
        action="store",
        dest="model_path",
        help="(required) path to SBML model with fbc information",
    )
    parser.add_option(
        "-p",
        "--path",
        action="store",
        dest="output_path",
        help="(required) directory path to write output files to",
    )
    parser.add_option(
        "-c",
        "--curator",
        action="store",
        dest="curator",
        help="(optional) curator tool to create reference files: Select from "
        "['cobrapy', 'cameo', 'all']",
    )
    parser.add_option(
        "-o",
        "--objective",
        action="store",
        dest="objective",
        help="(optional) objective to use in optimization, "
        "defaults to active objective",
    )
    parser.add_option(
        "-r",
        "--reference",
        action="store",
        dest="reference_path",
        help="(optional) directory path to reference output (comparison is performed)",
    )

    f = Figlet(font="slant")
    print(f.renderText("FBC CURATION"))
    print(f"\tVersion {__version__} (https://github.com/matthiaskoenig/fbc_curation)")
    print(f"\tCitation {__citation__}\n")

    options, args = parser.parse_args()

    def _parser_message(text):
        print(text)
        parser.print_help()
        sys.exit(1)

    if not options.model_path:
        _parser_message("Required argument '--model' missing")
    if not options.output_path:
        _parser_message("Required argument '--path' missing")

    model_path = Path(options.model_path)
    if not model_path.exists():
        _parser_message(
            f"--model '{options.model_path}' does not exist, ensure valid model path."
        )

    output_path = Path(options.output_path)
    if not output_path.exists():
        logger.warning(
            f"output path --path '{output_path}' does not exist, path will be created."
        )

    if not options.reference_path:
        reference_path = None
    else:
        reference_path = Path(options.reference_path)
        if not reference_path.exists():
            _parser_message(
                f"--reference '{options.reference_path}' does not exist, ensure "
                f"valid reference path."
            )

    if not options.curator:
        options.curator = "all"
    supported_curators = ["cameo", "cobrapy", "all"]
    if options.curator not in supported_curators:
        _parser_message(
            f"--curator '{options.curator}' is not supported, use one of the supported "
            f"curators: {supported_curators}"
        )

    obj_info = Curator.read_objective_information(model_path)
    if not options.objective:
        objective_id = obj_info.active_objective
    else:
        objective_id = options.objective
        if objective_id not in obj_info.objective_ids:
            _parser_message(
                f"Objective --objective'{objective_id}' dose not exist "
                f"in SBML model objectives: "
                f"'{obj_info.objective_ids}'"
            )
        elif not objective_id == obj_info.active_objective:
            _parser_message(
                f"Only active_objective supported in cobrapy, use "
                f"--objective {obj_info.active_objective}"
            )

    # Perform imports here to avoid messages above parser messages

    if options.curator == "cobrapy":
        from fbc_curation.curator.cobrapy_curator import CuratorCobrapy

        curator_classes = [CuratorCobrapy]
        curator_keys = ["cobrapy"]
    elif options.curator == "cameo":
        from fbc_curation.curator.cameo_curator import CuratorCameo

        curator_classes = [CuratorCameo]
        curator_keys = ["cameo"]
    elif options.curator == "all":
        from fbc_curation.curator.cameo_curator import CuratorCameo
        from fbc_curation.curator.cobrapy_curator import CuratorCobrapy

        curator_classes = [CuratorCobrapy, CuratorCameo]
        curator_keys = ["cobrapy", "cameo"]

    # Reading reference solution
    res_dict = {}
    if reference_path:
        reference_results = CuratorResults.read_results(reference_path)
        res_dict["reference"] = reference_results

    for k, curator_class in enumerate(curator_classes):
        key = curator_keys[k]
        curator = curator_class(model_path=model_path, objective_id=objective_id)
        results = curator.run()  # type: CuratorResults
        results.write_results(output_path / key)
        res_dict[key] = CuratorResults.read_results(output_path / key)

    # perform comparison
    if len(res_dict) > 1:
        CuratorResults.compare(res_dict)


if __name__ == "__main__":
    main()



"""Base class for all FBC curators."""
import logging
from collections import defaultdict, namedtuple
from pathlib import Path
from typing import Dict, List

import cobra
import libsbml
import pandas as pd
from cobra.io import read_sbml_model

from fbc_curation.curator.metadata import create_metadata
from fbc_curation.curator.results import CuratorResults


ObjectiveInformation = namedtuple(
    "ObjectiveInformation", "active_objective objective_ids"
)

logger = logging.getLogger(__name__)


class Curator:
    """Base class of all Curator implementations."""

    def __init__(self, model_path: Path, objective_id: str = None):
        """Create instance."""
        if not model_path.exists():
            raise ValueError(f"model_path does not exist: '{model_path}'")

        self.model_path = model_path
        self.active_objective, self.objective_ids = Curator.read_objective_information(
            model_path
        )

        if objective_id is None:
            logger.warning(
                f"No objective id provided, using the active objective: "
                f"{self.active_objective}"
            )
            self.objective_id = self.active_objective
        else:
            if objective_id not in self.objective_ids:
                logger.error(
                    f"objective id does not exist in model: '{self.objective_id}', "
                    f"using the active objective: {self.active_objective}"
                )
                self.objective_id = self.active_objective
            else:
                self.objective_id = objective_id

    def __str__(self):
        """Create string representation."""
        lines = [
            f"--- {self.__class__.__name__} ---",
            f"\tmodel_path: {self.model_path}",
            f"\tobjective_id: {self.objective_id}",
        ]
        return "\n".join(lines)

    def read_model(self):
        raise NotImplementedError

    def metadata(self) -> Dict:
        return create_metadata(path=self.model_path)

    def objective(self) -> pd.DataFrame:
        raise NotImplementedError

    def fva(self) -> pd.DataFrame:
        raise NotImplementedError

    def gene_deletion(self) -> pd.DataFrame:
        raise NotImplementedError

    def reaction_deletion(self) -> pd.DataFrame:
        raise NotImplementedError

    def run(self) -> CuratorResults:
        """Run the curator and stores the results."""

        print("-" * 80)
        self._print_header(f"{self.__class__.__name__}: metadata")
        metadata = self.metadata()

        self._print_header(f"{self.__class__.__name__}: objective")
        objective = self.objective()

        self._print_header(f"{self.__class__.__name__}: fva")
        fva = self.fva()

        self._print_header(f"{self.__class__.__name__}: gene_deletion")
        gene_deletion = self.gene_deletion()

        self._print_header(f"{self.__class__.__name__}: reaction_deletion")
        reaction_deletion = self.reaction_deletion()

        return CuratorResults(
            metadata=metadata,
            objective_id=self.objective_id,
            objective=objective,
            fva=fva,
            gene_deletion=gene_deletion,
            reaction_deletion=reaction_deletion,
        )

    @staticmethod
    def _print_header(title):
        print(f"* {title}")

    @staticmethod
    def knockout_reactions_for_genes(
        model_path: Path, genes=None
    ) -> Dict[str, List[str]]:
        """Calculate mapping of genes to affected reactions.

        Which reactions are knocked out by a given gene.
        A single gene knockout can affect multiple reactions.
        Uses GPR mappings.
        """
        model = read_sbml_model(str(model_path), f_replace={})  # type: cobra.core.Model
        if genes is None:
            genes = model.genes

        knockout_reactions = defaultdict(list)
        for reaction in model.reactions:  # type: cobra.core.Reaction
            gpr = reaction.gene_reaction_rule
            tree, gpr_genes = cobra.core.gene.parse_gpr(gpr)
            for gene in genes:  # type: cobra.core.Gene
                if gene.id not in gpr_genes:
                    gene_essential = False
                else:
                    # eval_gpr: True if the gene reaction rule is true with
                    # the given knockouts otherwise false
                    gene_essential = not cobra.core.gene.eval_gpr(
                        tree, knockouts={gene.id}
                    )
                if gene_essential:
                    knockout_reactions[gene.id].append(reaction.id)

        # from pprint import pprint
        # pprint(knockout_reactions)
        return knockout_reactions

    @staticmethod
    def read_objective_information(model_path: Path) -> ObjectiveInformation:
        """Read objective information from SBML file structure."""
        # read objective information from sbml (multiple objectives)
        doc = libsbml.readSBMLFromFile(str(model_path))  # type: libsbml.SBMLDocument
        model = doc.getModel()  # type: libsbml.Model
        fbc_model = model.getPlugin("fbc")  # type: libsbml.FbcModelPlugin
        if fbc_model is None:
            # model is an old SBML model without fbc information (use cobra default)
            # problems with the automatic up-conversions
            active_objective = "obj"
            objective_ids = ["obj"]
        else:
            active_objective = fbc_model.getActiveObjective().getId()
            objective_ids = []
            for objective in fbc_model.getListOfObjectives():  # type: libsbml.Objective
                objective_ids.append(objective.getId())

        if len(objective_ids) > 1:
            logger.warning(
                f"Multiple objectives exist in SBML-fbc ({objective_ids}), "
                f"only active objective '{active_objective}' results "
                f"are reported"
            )
        return ObjectiveInformation(
            active_objective=active_objective, objective_ids=objective_ids
        )


if __name__ == "__main__":
    from fbc_curation import EXAMPLE_PATH

    model_path = EXAMPLE_PATH / "models" / "e_coli_core.xml"
    Curator.knockout_reactions_for_genes(model_path=model_path)



"""Create curation information for example models."""
from pathlib import Path
from typing import Dict

from fbc_curation import EXAMPLE_PATH
from fbc_curation.curator import Curator, CuratorResults
from fbc_curation.curator.cameo_curator import CuratorCameo
from fbc_curation.curator.cobrapy_curator import CuratorCobrapy


def run_examples(results_path: Path = EXAMPLE_PATH / "results"):
    """Run all examples."""
    example_ecoli_core(results_path=results_path / "e_coli_core")
    example_iJR904(results_path=results_path / "iJR904")
    example_iAB_AMO1410_SARS(results_path=results_path / "iAB_AMO1410_SARS")


def example_ecoli_core(results_path: Path) -> Dict:
    """Create example files for ecoli core."""
    return _run_example(
        EXAMPLE_PATH / "models" / "e_coli_core.xml", results_path=results_path
    )


def example_iJR904(results_path: Path) -> Dict:
    """Create example files for iJR904."""
    return _run_example(
        EXAMPLE_PATH / "models" / "iJR904.xml.gz", results_path=results_path
    )


def example_iAB_AMO1410_SARS(results_path: Path) -> Dict:
    """Create example files for iAB_AMO1410_SARS."""
    return _run_example(
        EXAMPLE_PATH / "models" / "iAB_AMO1410_SARS-CoV-2.xml",
        results_path=results_path,
    )


def _run_example(model_path: Path, results_path: Path) -> Dict:
    print("#" * 80)
    print(model_path)
    print("#" * 80)
    obj_info = Curator.read_objective_information(model_path)

    curator_keys = ["cobrapy", "cameo"]
    for k, curator_class in enumerate([CuratorCobrapy, CuratorCameo]):
        curator = curator_class(
            model_path=model_path, objective_id=obj_info.active_objective
        )
        results = curator.run()  # type: CuratorResults
        results.write_results(results_path / curator_keys[k])

    all_results = {}
    for curator_key in curator_keys:
        all_results[curator_key] = CuratorResults.read_results(
            path_in=results_path / curator_key
        )

    # comparison
    valid = [r.validate() for r in all_results.values()]
    equal = CuratorResults.compare(all_results)
    info = {
        "model_path": model_path,
        "valid": valid,
        "equal": equal,
    }
    print(info)
    return info


if __name__ == "__main__":
    # run_examples()
    example_ecoli_core(results_path=EXAMPLE_PATH / "results" / "e_coli_core")



"""Create metadata information."""
import hashlib
import os
import platform
from pathlib import Path

from fbc_curation import __citation__, __software__, __version__


def create_metadata(path: Path):
    """Create curation metadata for given model."""
    d = {}
    d["software.name"] = __software__
    d["software.version"] = __version__
    d["software.url"] = __citation__
    d["environment"] = f"{os.name}, {platform.system()}, {platform.release()}"
    d["model.filename"] = path.name
    d["model.md5"] = md5_for_path(path)
    d["solver.name"] = None
    d["solver.version"] = None

    return d


def md5_for_path(path: Path):
    """Calculate MD5 of file content."""

    # Open,close, read file and calculate MD5 on its contents
    with open(path, "rb") as f_check:
        # read contents of the file
        data = f_check.read()
        # pipe contents of the file through
        return hashlib.md5(data).hexdigest()


def filename_for_path(path: Path):
    """Return filename of path."""
    return path.name


if __name__ == "__main__":

    from fbc_curation import EXAMPLE_PATH

    ecoli_path = EXAMPLE_PATH / "models" / "e_coli_core.xml"
    d = create_metadata(path=ecoli_path)
    print(d)



"""Test the fbc_curation on the examples."""
from pathlib import Path

import pytest

from fbc_curation import examples
from fbc_curation.constants import CuratorConstants


def _check_tmp_path(path):
    for curator_key in ["cobrapy", "cameo"]:
        assert Path.exists(path / curator_key / CuratorConstants.OBJECTIVE_FILENAME)
        assert Path.exists(path / curator_key / CuratorConstants.FVA_FILENAME)
        assert Path.exists(
            path / curator_key / CuratorConstants.REACTION_DELETION_FILENAME
        )
        assert Path.exists(path / curator_key / CuratorConstants.GENE_DELETION_FILENAME)


def _check_example_results(res):
    assert res["valid"] == [True, True]
    assert res["equal"]


def test_e_coli_core(tmp_path):
    """Test fbc_curation."""
    res = examples.example_ecoli_core(tmp_path)
    _check_tmp_path(tmp_path)
    _check_example_results(res)


def test_iJR904(tmp_path):
    """Test fbc_curation."""
    res = examples.example_iJR904(tmp_path)
    _check_tmp_path(tmp_path)
    _check_example_results(res)


@pytest.mark.skip(reason="Failing test")
def test_iAB_AMO1410_SARS(tmp_path):
    """Test fbc_curation."""
    res = examples.example_iAB_AMO1410_SARS(tmp_path)
    _check_tmp_path(tmp_path)
    _check_example_results(res)



"""Curate fbc with cameo."""

import logging
from pathlib import Path
from typing import Dict

import pandas as pd
from cameo import __version__ as cameo_version
from cameo import fba
from cameo.flux_analysis.analysis import (
    FluxVariabilityResult,
    flux_variability_analysis,
)
from cobra.core import Model
from cobra.io import read_sbml_model

from fbc_curation.constants import CuratorConstants
from fbc_curation.curator import Curator


logger = logging.getLogger(__name__)


class CuratorCameo(Curator):
    """FBC curator based on cameo.

    Cameo is a high-level python library developed to aid the strain design
    process in metabolic engineering projects. The library provides a modular
    framework of simulation and strain design methods that targets developers
    that want to develop new design algorithms and custom analysis workflows.
    Furthermore, it exposes a high-level API to users that just want to
    compute promising strain designs.

    https://pythonhosted.org/cameo/
    """

    def __init__(self, model_path: Path, objective_id: str = None):
        """Create instance."""
        Curator.__init__(self, model_path=model_path, objective_id=objective_id)

    def read_model(self) -> Model:
        return read_sbml_model(str(self.model_path), f_replace={})

    def metadata(self) -> Dict:
        """Create metadata dictionary."""
        d = super().metadata()
        d["solver.name"] = "cameo (glpk)"
        d["solver.version"] = f"{cameo_version}"
        return d

    def objective(self) -> pd.DataFrame:
        model = self.read_model()
        try:
            # fbc optimization
            result = fba(model)
            value = result.objective_value
            status = CuratorConstants.STATUS_OPTIMAL
        except Exception as e:
            logger.error(f"{e}")
            value = CuratorConstants.VALUE_INFEASIBLE
            status = CuratorConstants.STATUS_INFEASIBLE

        return pd.DataFrame(
            {
                "model": self.model_path.name,
                "objective": [self.objective_id],
                "status": [status],
                "value": [value],
            }
        )

    def fva(self) -> pd.DataFrame:
        model = self.read_model()
        result = fba(model)
        fluxes = result.fluxes
        try:
            fva_result = flux_variability_analysis(
                model, reactions=model.reactions, fraction_of_optimum=1.0
            )  # type: FluxVariabilityResult
            df = fva_result.data_frame
            df_out = pd.DataFrame(
                {
                    "model": self.model_path.name,
                    "objective": self.objective_id,
                    "reaction": df.index,
                    "flux": fluxes,
                    "status": CuratorConstants.STATUS_OPTIMAL,
                    "minimum": df.lower_bound,
                    "maximum": df.upper_bound,
                }
            )
        except Exception as e:
            logger.error(f"{e}")
            df_out = pd.DataFrame(
                {
                    "model": self.model_path.name,
                    "objective": self.objective_id,
                    "reaction": [r.id for r in model.reactions],
                    "flux": fluxes,
                    "status": CuratorConstants.STATUS_INFEASIBLE,
                    "minimum": CuratorConstants.VALUE_INFEASIBLE,
                    "maximum": CuratorConstants.VALUE_INFEASIBLE,
                }
            )
        return df_out

    def gene_deletion(self) -> pd.DataFrame:
        model = self.read_model()
        gene_status = []
        gene_values = []

        knockout_reactions = self.knockout_reactions_for_genes(self.model_path)

        for gene in model.genes:
            reaction_bounds = dict()
            # knockout all reactions affected by gene by setting bounds zero
            for rid in knockout_reactions[gene.id]:
                reaction = model.reactions.get_by_id(rid)
                reaction_bounds[reaction.id] = (
                    reaction.lower_bound,
                    reaction.upper_bound,
                )
                reaction.bounds = (0, 0)
            try:
                # run fba
                result = fba(model)
                value = result.objective_value
                status = CuratorConstants.STATUS_OPTIMAL
            except Exception as e:
                logger.error(f"{e}")
                value = CuratorConstants.VALUE_INFEASIBLE
                status = CuratorConstants.STATUS_INFEASIBLE
            gene_status.append(status)
            gene_values.append(value)

            # restore bounds
            for rid, bounds in reaction_bounds.items():
                model.reactions.get_by_id(rid).bounds = bounds[:]

        return pd.DataFrame(
            {
                "model": self.model_path.name,
                "objective": self.objective_id,
                "gene": [gene.id for gene in model.genes],
                "status": gene_status,
                "value": gene_values,
            }
        )

    def reaction_deletion(self) -> pd.DataFrame:
        model = self.read_model()
        reaction_status = []
        reaction_values = []

        for reaction in model.reactions:
            reaction_bounds = (reaction.lower_bound, reaction.upper_bound)
            reaction.bounds = (0, 0)

            # run fba
            try:
                result = fba(model)
                value = result.objective_value
                status = CuratorConstants.STATUS_OPTIMAL
            except Exception as e:
                logger.error(f"{e}")
                value = CuratorConstants.VALUE_INFEASIBLE
                status = CuratorConstants.STATUS_INFEASIBLE

            reaction_status.append(status)
            reaction_values.append(value)

            # restore bounds
            reaction.bounds = reaction_bounds[:]

        return pd.DataFrame(
            {
                "model": self.model_path.name,
                "objective": self.objective_id,
                "reaction": [r.id for r in model.reactions],
                "status": reaction_status,
                "value": reaction_values,
            }
        )



"""Reused constants."""


class CuratorConstants:
    """Class storing constants for curation and file format."""

    # keys of outputs
    METADATA_KEY = "metadata"
    OBJECTIVE_KEY = "objective"
    FVA_KEY = "fva"
    GENE_DELETION_KEY = "gene_deletion"
    REACTION_DELETION_KEY = "reaction_deletion"
    KEYS = [
        OBJECTIVE_KEY,
        FVA_KEY,
        GENE_DELETION_KEY,
        REACTION_DELETION_KEY,
    ]

    # output filenames
    METADATA_FILENAME = "metadata.json"
    OBJECTIVE_FILENAME = f"01_{OBJECTIVE_KEY}.tsv"
    FVA_FILENAME = f"02_{FVA_KEY}.tsv"
    GENE_DELETION_FILENAME = f"03_{GENE_DELETION_KEY}.tsv"
    REACTION_DELETION_FILENAME = f"04_{REACTION_DELETION_KEY}.tsv"
    FILENAMES = [
        OBJECTIVE_FILENAME,
        FVA_FILENAME,
        GENE_DELETION_FILENAME,
        REACTION_DELETION_FILENAME,
    ]

    # fields
    OBJECTIVE_FIELDS = ["model", "objective", "status", "value"]
    FVA_FIELDS = [
        "model",
        "objective",
        "reaction",
        "flux",
        "status",
        "minimum",
        "maximum",
    ]
    GENE_DELETION_FIELDS = ["model", "objective", "gene", "status", "value"]
    REACTION_DELETION_FIELDS = ["model", "objective", "reaction", "status", "value"]

    # status codes
    STATUS_OPTIMAL = "optimal"
    STATUS_INFEASIBLE = "infeasible"
    STATUS_CODES = [STATUS_OPTIMAL, STATUS_INFEASIBLE]

    # special settings for comparison
    VALUE_INFEASIBLE = ""  # pd.NA
    NUM_DECIMALS = 6  # decimals to write in the solution



"""Provide cobrapy fbc curator."""

import logging
from pathlib import Path
from typing import Dict

import pandas as pd
from cobra import __version__ as cobra_version
from cobra.core import Model
from cobra.exceptions import OptimizationError
from cobra.flux_analysis import (
    flux_variability_analysis,
    single_gene_deletion,
    single_reaction_deletion,
)
from cobra.io import read_sbml_model

from fbc_curation.constants import CuratorConstants
from fbc_curation.curator import Curator


logger = logging.getLogger(__file__)


class CuratorCobrapy(Curator):
    """FBC curator based on cobrapy."""

    def __init__(self, model_path: Path, objective_id: str = None):
        """Create instance."""
        super().__init__(model_path=model_path, objective_id=objective_id)

    def read_model(self) -> Model:
        """Read the model."""
        return read_sbml_model(str(self.model_path), f_replace={})

    def metadata(self) -> Dict:
        """Create metadata dictionary."""
        d = super().metadata()
        d["solver.name"] = "cobrapy (glpk)"
        d["solver.version"] = f"{cobra_version}"
        return d

    def objective(self) -> pd.DataFrame:
        """Create pandas DataFrame with objective value.

        see https://cobrapy.readthedocs.io/en/latest/simulating.html
        """
        model = self.read_model()
        try:
            # fbc optimization
            solution = model.optimize()
            value = solution.objective_value
            status = CuratorConstants.STATUS_OPTIMAL
        except OptimizationError as e:
            logger.error(f"{e}")
            value = CuratorConstants.VALUE_INFEASIBLE
            status = CuratorConstants.STATUS_INFEASIBLE

        return pd.DataFrame(
            {
                "model": self.model_path.name,
                "objective": [self.objective_id],
                "status": [status],
                "value": [value],
            }
        )

    def fva(self) -> pd.DataFrame:
        """Create DataFrame file with minimum and maximum value of FVA.

        Runs flux variability analysis.
        see https://cobrapy.readthedocs.io/en/latest/simulating.html#Running-FVA
        """
        model = self.read_model()
        solution = model.optimize()
        fluxes = solution.fluxes
        try:
            df = flux_variability_analysis(
                model, model.reactions, fraction_of_optimum=1.0
            )
            df_out = pd.DataFrame(
                {
                    "model": self.model_path.name,
                    "objective": self.objective_id,
                    "reaction": df.index,
                    "flux": fluxes,
                    "status": CuratorConstants.STATUS_OPTIMAL,
                    "minimum": df.minimum,
                    "maximum": df.maximum,
                }
            )
        except OptimizationError as e:
            logger.error(f"{e}")
            df_out = pd.DataFrame(
                {
                    "model": self.model_path.name,
                    "objective": self.objective_id,
                    "reaction": [r.id for r in model.reactions],
                    "flux": fluxes,
                    "status": CuratorConstants.STATUS_INFEASIBLE,
                    "minimum": CuratorConstants.VALUE_INFEASIBLE,
                    "maximum": CuratorConstants.VALUE_INFEASIBLE,
                }
            )

        return df_out

    def gene_deletion(self) -> pd.DataFrame:
        """Create pd.DataFrame with results of gene deletion.

        https://cobrapy.readthedocs.io/en/latest/deletions.html
        :return: pandas.DataFrame
        """
        model = self.read_model()
        df = single_gene_deletion(model, model.genes)
        print(df)
        return pd.DataFrame(
            {
                "model": self.model_path.name,
                "objective": self.objective_id,
                "gene": [set(ids).pop() for ids in df.ids],
                "status": df.status,
                "value": df.growth,
            }
        )

    def reaction_deletion(self) -> pd.DataFrame:
        """Create pd.DataFramewith results of reaction deletion.

        https://cobrapy.readthedocs.io/en/latest/deletions.html
        :return: pandas.
        """
        model = self.read_model()
        df = single_reaction_deletion(model, model.reactions)
        return pd.DataFrame(
            {
                "model": self.model_path.name,
                "objective": self.objective_id,
                "reaction": [set(ids).pop() for ids in df.ids],
                "status": df.status,
                "value": df.growth,
            }
        )



"""Test fbc_curation command line script."""

import sys

import pytest

from fbc_curation import EXAMPLE_PATH, curation
from fbc_curation.constants import CuratorConstants


@pytest.mark.parametrize("filename", ["e_coli_core.xml", "iJR904.xml.gz"])
def test_curation1(monkeypatch, tmp_path, filename):
    """First example via command line tool.

    curation --model ../examples/models/e_coli_core.xml
    --path ../examples/results/e_coli_core
    """
    with monkeypatch.context() as m:
        args = [
            "curation",
            "--model",
            f"{EXAMPLE_PATH / 'models' / filename}",
            "--path",
            str(tmp_path),
        ]
        m.setattr(sys, "argv", args)
        curation.main()
        for out_fname in CuratorConstants.FILENAMES:
            assert (tmp_path / "cobrapy" / out_fname).exists()
            assert (tmp_path / "cameo" / out_fname).exists()


def test_curation2(monkeypatch, tmp_path):
    """Second example via command line tool.

    curation --model examples/models/e_coli_core.xml
    --path examples/results/e_coli_core
    --reference ../examples/results/e_coli_core/cobrapy
    """
    with monkeypatch.context() as m:
        args = [
            "curation",
            "--model",
            f"{EXAMPLE_PATH / 'models' / 'e_coli_core.xml'}",
            "--path",
            str(tmp_path),
            "--reference",
            f"{EXAMPLE_PATH / 'results' / 'e_coli_core' / 'cobrapy'}",
        ]
        m.setattr(sys, "argv", args)
        curation.main()
        for out_fname in CuratorConstants.FILENAMES:
            assert (tmp_path / "cobrapy" / out_fname).exists()
            assert (tmp_path / "cameo" / out_fname).exists()



"""Testing result."""

import libsbml
import pandas as pd

from fbc_curation import EXAMPLE_PATH
from fbc_curation.constants import CuratorConstants
from fbc_curation.curator import CuratorResults
from fbc_curation.curator.cobrapy_curator import CuratorCobrapy


model_path = EXAMPLE_PATH / "models" / "e_coli_core.xml"
curator = CuratorCobrapy(model_path=model_path)
results = curator.run()
doc = libsbml.readSBMLFromFile(str(model_path))  # type: libsbml.SBMLDocument
model = doc.getModel()  # type: libsbml.Model


def _check_objective(df):
    """Check objective DataFrame."""
    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert len(df.columns) == len(CuratorConstants.OBJECTIVE_FIELDS)
    for field in CuratorConstants.OBJECTIVE_FIELDS:
        assert field in df.columns
    for k, field in enumerate(CuratorConstants.OBJECTIVE_FIELDS):
        assert df.columns[k] == field

    obj_value = df["value"].values[0]
    assert obj_value > 0

    status_codes = df.status.unique()
    assert len(status_codes) <= 2
    assert "optimal" in status_codes


def test_objective():
    """Check objective."""
    df = results.objective
    _check_objective(df)


def _check_fva(df):
    """Check FVA."""
    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert len(df.columns) == len(CuratorConstants.FVA_FIELDS)
    for field in CuratorConstants.FVA_FIELDS:
        assert field in df.columns
    for k, field in enumerate(CuratorConstants.FVA_FIELDS):
        assert df.columns[k] == field

    assert len(df) == model.getNumReactions()

    status_codes = df.status.unique()
    assert len(status_codes) <= 2
    assert "optimal" in status_codes


def test_fva():
    """Check FVA DataFrame."""
    df = results.fva
    _check_fva(df)


def _check_gene_deletion(df):
    """Check gene deletion."""
    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert len(df.columns) == len(CuratorConstants.GENE_DELETION_FIELDS)
    for field in CuratorConstants.GENE_DELETION_FIELDS:
        assert field in df.columns
    for k, field in enumerate(CuratorConstants.GENE_DELETION_FIELDS):
        assert df.columns[k] == field

    fbc_model = model.getPlugin("fbc")  # type: libsbml.FbcModelPlugin
    assert len(df) == fbc_model.getNumGeneProducts()

    status_codes = df.status.unique()
    assert len(status_codes) <= 2
    assert "optimal" in status_codes


def test_gene_deletion():
    """Check gene deletion."""
    df = results.gene_deletion
    _check_gene_deletion(df)


def _check_reaction_deletion(df):
    """Check reaction deletion."""
    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert len(df.columns) == len(CuratorConstants.REACTION_DELETION_FIELDS)
    for field in CuratorConstants.REACTION_DELETION_FIELDS:
        assert field in df.columns
    for k, field in enumerate(CuratorConstants.REACTION_DELETION_FIELDS):
        assert df.columns[k] == field

    assert len(df) == model.getNumReactions()

    status_codes = df.status.unique()
    assert len(status_codes) <= 2
    assert "optimal" in status_codes


def test_reaction_deletion(tmp_path):
    """Check reaction deletion."""
    df = results.reaction_deletion
    _check_reaction_deletion(df)


def test_read_write_check1(tmp_path):
    """Test reading/writing of results."""
    results.write_results(tmp_path)
    results2 = CuratorResults.read_results(tmp_path)
    assert CuratorResults.compare({"res1": results, "res2": results2})


def test_read_write_check2(tmp_path):
    """Test reading/writing of results."""
    results.write_results(tmp_path)
    results2 = CuratorResults.read_results(tmp_path)
    assert results2.validate_objective()
    assert results2.validate_fva()
    assert results2.validate_gene_deletion()
    assert results2.validate_reaction_deletion()
    assert results2.validate()
